package org.mmc.model.service;

import org.mmc.model.domain.*;
import org.mmc.model.filters.ProbesFilter;
import org.mmc.model.persistence.dao.ProbesDao;
import org.mmc.model.location.TreeNode;
import org.mmc.model.print.ColumnProps;
import org.mmc.model.print.ProbesToXMLConvertor;
import org.mmc.model.print.NameValuePair;
import org.mmc.model.transformers.TransformationManager;
import org.mmc.utils.DateTimeUtils;

import java.util.List;
import java.util.Date;
import java.util.ArrayList;
import java.io.InputStream;

/**
 * Created on: 2007-10-9
 *
 * @author Stoyo Venkov
 */
public class ProbesDaoService implements ProbesService {
    private ProbesDao dao;

    private PatientsService patientsService;
    private AnalysisRequestsService requestsService;
    private ContainerAndGroupsService containersService;
    private NomenclaturesService nomenclaturesService;

    public ProbesDao getDao() {
        return dao;
    }

    public void setDao(ProbesDao dao) {
        this.dao = dao;
    }

    public PatientsService getPatientsService() {
        return patientsService;
    }

    public void setPatientsService(PatientsService patientsService) {
        this.patientsService = patientsService;
    }

    public AnalysisRequestsService getRequestsService() {
        return requestsService;
    }

    public void setRequestsService(AnalysisRequestsService requestsService) {
        this.requestsService = requestsService;
    }

    public ContainerAndGroupsService getContainersService() {
        return containersService;
    }

    public void setContainersService(ContainerAndGroupsService containersService) {
        this.containersService = containersService;
    }

    public NomenclaturesService getNomenclaturesService() {
        return nomenclaturesService;
    }

    public void setNomenclaturesService(NomenclaturesService nomenclaturesService) {
        this.nomenclaturesService = nomenclaturesService;
    }

    public Probe getProbeById(Long id) {
        Probe result = dao.getProbeById(id);
        result.setSteps(dao.getProbeSteps(result));
        return result;
    }

    public List getProbesByFilter(ProbesFilter filter) {
        return dao.getProbesByFilter(filter);
    }

    public Probe updateProbe(Probe probe) {
        return dao.updateProbe(probe);
    }

    public Probe registerProbe(Probe probe) {
        return dao.saveProbe(probe);
    }

    public Probe registerProbe(Long userId, Long patientId, Long requestId, List criteria,
                               List projects, Integer termInDays, Long probeTypeId, Long tubeId) {
        Probe result = new Probe();

        User creator = nomenclaturesService.getUsersService().getUserById(userId.longValue());

        result.setRegistrationDate(new Date());
        result.setTermInDays(termInDays);
        result.setType(nomenclaturesService.getProbeTypesService().getProbeTypeById(probeTypeId.longValue()));
        result.setCreator(creator);

        Patient patient = patientsService.getPatientById(patientId);
        result.setPatient(patient);
        AnalysisRequest request = requestsService.getRequestById(requestId);
        result.setRequest(request);

        //init projects list
        List probeProjects = new ArrayList(projects.size());
        for (int i = 0; i < projects.size(); i++) {
            Project project = (Project) projects.get(i);
            String code = project.getCode() + "_" +
                    patient.getUniqueNumber() + "_" +
                    DateTimeUtils.formatDateNoDelim(result.getRegistrationDate());
            ProbeProject pp = new ProbeProject();
            pp.setProbeCode(code);
            pp.setProject(project);
            probeProjects.add(pp);
        }
        result.setProbeProjects(probeProjects);

        //init criteria list
        List probeCriteria = new ArrayList(criteria.size());
        for (int i = 0; i < criteria.size(); i++) {
            Criterion criterion = (Criterion) criteria.get(i);
            ProbeCriterion pc = new ProbeCriterion();
            pc.setCriterion(criterion);
            probeCriteria.add(pc);
        }
        result.setProbeCriteria(probeCriteria);

        //save probe
        result = registerProbe(result);

        //register first step
        Step step = registerFirstStep(result.getId(), creator);

        //register location
        registerLocation(result.getId(), step.getId(), null, tubeId, null, null, null, null);

        return result;
    }

    public Step registerStep(Long userId, Long prevStepId, Long probeId, Long projectId, Long kindId, String notes) {
        User creator = nomenclaturesService.getUsersService().getUserById(userId.longValue());
        Date dt = new Date();

        if(prevStepId != null) {
            Step prevStep = dao.getStepById(prevStepId);
            if(prevStep.getFinishDate() == null) {
                prevStep.setFinishDate(dt);
                prevStep.setSuccessful(Boolean.TRUE);
                dao.updateStep(prevStep);
            }
        }

        Step step = new Step();
        step.setCreator(creator);
        step.setProbeId(probeId);
        step.setProjectId(projectId);
        step.setNotes(notes);
        step.setParentStepId(prevStepId);
        step.setStartDate(dt);
        step.setNomenclatureStep(nomenclaturesService.getStepsService().getProcessStepById(kindId));

        step = dao.saveStep(step);

        return step;
    }

    public Location registerLocation(Long probeId, Long stepId, Long prevLocationId,
                                     Long tubeId, Long rackId, Long plateId, Long batchGroupId, String coordinates) {

        Location location = new Location();
        location.setRegistrationDate(new Date());
        location.setProbeId(probeId);
        location.setStepId(stepId);
        if(tubeId != null) {
            location.setTube(containersService.getTubeById(tubeId));
        } else {
            if(rackId != null) {
                location.setRack(containersService.getRackById(rackId));
            } else if(plateId != null) {
                location.setPlate(containersService.getPlateById(plateId));
            }
            if(batchGroupId != null) {
                location.setGroup(containersService.getBatchGroupById(batchGroupId));
            }
            location.setCoordinates(coordinates);
        }

        location = dao.saveLocation(location);

        if(prevLocationId != null) {
            LocationChange lc = new LocationChange();
            lc.setSrcId(prevLocationId);
            lc.setDestId(location.getId());
            dao.registerMovement(lc);
        }

        return location;
    }

    public String getLocationsTable(Long stepId) {
        TreeNode root = getLocationsTree(stepId);
        StringBuffer result = convertTreeToTable(root);
        return result.toString();
    }

    public Probe cancelProbe(Probe probe, String reason) {
        probe.setCancelationReason(reason);
        probe.setCanceled(Boolean.TRUE);
        return updateProbe(probe);
    }

    public Probe markProbeAsSent(Probe probe) {
        probe.setResultSent(Boolean.TRUE);
        return updateProbe(probe);
    }

    public Probe completeProbe(Probe probe) {
        probe.setFinished(Boolean.TRUE);
        probe.setFinishDate(new Date());
        return updateProbe(probe);
    }

    public byte[] executeProjectReport(Long projectId, Date regFromDate, Date regToDate) {
        ProbesFilter filter = new ProbesFilter(0);
        filter.setProjectId(projectId);
        filter.setRegFromDate(regFromDate);
        filter.setRegToDate(regToDate);

        ArrayList columnProperties = new ArrayList(6);
        columnProperties.add(new ColumnProps("id", ". ", 20));
        columnProperties.add(new ColumnProps("regDateStr", " ", 15));
        columnProperties.add(new ColumnProps("termInDays", "   ", 10));
        columnProperties.add(new ColumnProps("finished", "", 10));
        columnProperties.add(new ColumnProps("canceled", "", 10));
        columnProperties.add(new ColumnProps("type", "", 10));
        columnProperties.add(new ColumnProps("creator", " ", 25));

        ArrayList params = new ArrayList(2);
        if(projectId != null) {
            params.add(new NameValuePair("",
                    nomenclaturesService.getProjectsService().getProjectById(projectId.longValue()).toString()));
        }
        if(regFromDate != null) {
            params.add(new NameValuePair(" ", DateTimeUtils.formatDate(regFromDate)));
        }
        if(regToDate != null) {
            params.add(new NameValuePair(" ", DateTimeUtils.formatDate(regToDate)));
        }

        List probes = dao.getAllProbesByFilter(filter);
        if(probes != null) {
            ProbesToXMLConvertor convertor =
                    new ProbesToXMLConvertor("        ",
                            params, columnProperties, probes, "");

            InputStream xml = convertor.createXML();
            TransformationManager manager = new TransformationManager();
            return manager.transformToPDF(xml);
        } else {
            return null;
        }
    }

    public byte[] executePatientReport(Long patientId, Date regToDate) {
        ProbesFilter filter = new ProbesFilter(0);
        filter.setPatientId(patientId);
        filter.setRegToDate(regToDate);

        ArrayList columnProperties = new ArrayList(8);
        columnProperties.add(new ColumnProps("id", ". ", 15));
        columnProperties.add(new ColumnProps("regDateStr", " ", 11));
        columnProperties.add(new ColumnProps("termInDays", "   ", 10));
        columnProperties.add(new ColumnProps("finalDateStr", "  ", 10));
        columnProperties.add(new ColumnProps("finished", "", 10));
        columnProperties.add(new ColumnProps("canceled", "", 10));
        columnProperties.add(new ColumnProps("type", "", 10));
        columnProperties.add(new ColumnProps("creator", " ", 25));

        ArrayList params = new ArrayList(2);
        if(patientId != null) {
            params.add(new NameValuePair("", patientsService.getPatientById(patientId).toString()));
        }
        if(regToDate != null) {
            params.add(new NameValuePair(" ", DateTimeUtils.formatDate(regToDate)));
        }

        List probes = dao.getAllProbesByFilter(filter);
        if(probes != null) {
            ProbesToXMLConvertor convertor =
                    new ProbesToXMLConvertor("      ",
                            params, columnProperties, probes, "");

            InputStream xml = convertor.createXML();
            TransformationManager manager = new TransformationManager();
            return manager.transformToPDF(xml);
        } else {
            return null;
        }
    }

    public byte[] executeExpiredReport(Long creatorId, Long projectId) {
        ProbesFilter filter = new ProbesFilter(0);
        filter.setCreatorId(creatorId);
        filter.setProjectId(projectId);
        filter.setTerm(ProbesFilter.EXPIRED_PROBES);

        ArrayList columnProperties = new ArrayList(6);
        columnProperties.add(new ColumnProps("id", ". ", 20));
        columnProperties.add(new ColumnProps("regDateStr", " ", 15));
        columnProperties.add(new ColumnProps("termInDays", "   ", 12));
        columnProperties.add(new ColumnProps("daysPastExpiration", "   ", 12));
        columnProperties.add(new ColumnProps("type", "", 11));
        columnProperties.add(new ColumnProps("creator", " ", 30));

        ArrayList params = new ArrayList(2);
        if(creatorId != null) {
            params.add(new NameValuePair(" ",
                    nomenclaturesService.getUsersService().getUserById(creatorId.longValue()).toString()));
        }
        if(projectId != null) {
            params.add(new NameValuePair("",
                    nomenclaturesService.getProjectsService().getProjectById(projectId.longValue()).toString()));
        }

        List probes = dao.getAllProbesByFilter(filter);
        if(probes != null) {
            ProbesToXMLConvertor convertor =
                    new ProbesToXMLConvertor("        ",
                            params, columnProperties, probes, "");

            InputStream xml = convertor.createXML();
            TransformationManager manager = new TransformationManager();
            return manager.transformToPDF(xml);
        } else {
            return null;
        }
    }

    public byte[] executePressingReport(Long creatorId, Long projectId) {
        ProbesFilter filter = new ProbesFilter(0);
        filter.setCreatorId(creatorId);
        filter.setProjectId(projectId);
        filter.setTerm(ProbesFilter.PRESSING_PROBES);

        ArrayList columnProperties = new ArrayList(6);
        columnProperties.add(new ColumnProps("id", ". ", 20));
        columnProperties.add(new ColumnProps("regDateStr", " ", 15));
        columnProperties.add(new ColumnProps("termInDays", "   ", 10));
        columnProperties.add(new ColumnProps("finalDateStr", "  ", 15));
        columnProperties.add(new ColumnProps("type", "", 10));
        columnProperties.add(new ColumnProps("creator", " ", 30));

        ArrayList params = new ArrayList(2);
        if(creatorId != null) {
            params.add(new NameValuePair(" ",
                    nomenclaturesService.getUsersService().getUserById(creatorId.longValue()).toString()));
        }
        if(projectId != null) {
            params.add(new NameValuePair("",
                    nomenclaturesService.getProjectsService().getProjectById(projectId.longValue()).toString()));
        }

        List probes = dao.getAllProbesByFilter(filter);
        if(probes != null) {
            ProbesToXMLConvertor convertor =
                    new ProbesToXMLConvertor("        ",
                            params, columnProperties, probes, "");

            InputStream xml = convertor.createXML();
            TransformationManager manager = new TransformationManager();
            return manager.transformToPDF(xml);
        } else {
            return null;
        }
    }

    private TreeNode getLocationsTree(Long stepId) {
        Step step = dao.getStepById(stepId);
        Long probeId = step.getProbeId();
        Long projectId = step.getProjectId();
        Probe probe = getProbeById(probeId);
        Step firstStep = (Step) ((StepsInProject) probe.getSteps().get(0)).getSteps().get(0);
        Location firstLocation = dao.getFirstLocation(firstStep.getId());
        return createLocationsTree(firstLocation.getId(), projectId);
    }

    private StringBuffer convertTreeToTable(TreeNode node) {
        StringBuffer result = new StringBuffer("");
        result.append("<table class=\"locationTable\">");
        result.append("<tr>");
        result.append("<td align=\"left\" ");
        if(node.getElements() != null && node.getElements().size() > 0) {
            result.append("width=\"200\">");
            result.append("<a href=\"#\" onClick=\"doShowHide(").
                    append(node.getLocation().getId()).append(")\" class=\"same\" id=\"hrefId").
                    append(node.getLocation().getId()).
                    append("\">-</a>").
                    append("<br/>");
        } else {
            result.append("colspan=\"2\">");
        }
        result.append(": ").append(node.getLocation().getStep().getId()).append("<br/>");
        if(node.getLocation().getTube() != null) {
            result.append(": ").append(node.getLocation().getTube().getId()).append("<br/>");
            result.append(": ").append(node.getLocation().getTube().getLocation()).append("<br/>");
        } else if(node.getLocation().getRack() != null) {
            result.append(": ").append(node.getLocation().getRack().getId()).append("<br/>");
            result.append(": ").append(node.getLocation().getCoordinates()).append("<br/>");
            result.append(": ").append(node.getLocation().getRack().getLocation()).append("<br/>");
        } else if(node.getLocation().getPlate() != null) {
            result.append(": ").append(node.getLocation().getPlate().getId()).append("<br/>");
            result.append(": ").append(node.getLocation().getCoordinates()).append("<br/>");
            result.append(": ").append(node.getLocation().getPlate().getLocation()).append("<br/>");
        }
        if(node.getLocation().getGroup() != null) {
            result.append(": ").append(node.getLocation().getGroup().getId()).append("<br/>");
        }
        result.append("<input type=\"button\" class=\"buttonAuto\" value=\"\" onclick=\"doMove(").
                append(node.getLocation().getId()).
                append(")\">");
        result.append("</td>");
        if(node.getElements() != null && node.getElements().size() > 0) {
            result.append("<td>");
            result.append("<div id=\"divId").append(node.getLocation().getId()).append("\">");
            result.append("<table class=\"locationTable\">");
            for (int i = 0; i < node.getElements().size(); i++) {
                TreeNode treeNode = (TreeNode) node.getElements().get(i);
                result.append("<tr><td>").append(convertTreeToTable(treeNode)).append("</td></tr>");
            }
            result.append("</table>");
            result.append("</div>");
            result.append("<div id=\"emptyDivId").append(node.getLocation().getId()).append("\"").append("&nbsp;").append("</div>");
            result.append("</td>");
        }
        result.append("</tr>");
        result.append("</table>");
        return result;
    }

    private TreeNode createLocationsTree(Long locationId, Long projectId) {
        TreeNode result = new TreeNode();
        result.setLocation(dao.getLocationById(locationId));

        List children = dao.getChildrenLocations(locationId, projectId);
        if(children != null && children.size() > 0) {
            result.setElements(new ArrayList(children.size()));
            for (int i = 0; i < children.size(); i++) {
                Location location = (Location) children.get(i);
                result.getElements().add(createLocationsTree(location.getId(), projectId));
            }
        }

        return result;
    }

    private Step registerFirstStep(Long probeId, User creator) {
        Step step = new Step();
        step.setCreator(creator);
        step.setProbeId(probeId);
        step.setStartDate(new Date());
        step.setNomenclatureStep(nomenclaturesService.getStepsService().getFirstStep());
        step.setSuccessful(Boolean.TRUE);
        step.setFinishDate(new Date());
        return dao.saveStep(step);
    }
}
