package org.mmc.model.transformers;

import org.apache.fop.apps.Driver;
import org.apache.fop.apps.Options;
import org.apache.fop.apps.FOPException;
import org.apache.fop.messaging.MessageHandler;
import org.apache.fop.configuration.Configuration;
import org.apache.avalon.framework.logger.Log4JLogger;
import org.apache.log4j.Logger;
import org.mmc.utils.StringUtils;

import javax.xml.transform.*;
import javax.xml.transform.sax.SAXResult;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.util.Properties;

/**
 * Created on: 2007-10-18
 *
 * @author Stoyo Venkov
 */
public class FOPTransformer implements BaseTransformer {
    private static Logger log = Logger.getLogger(FOPTransformer.class.getName());

    private Driver fopDriver;// static

    /*This field defines the configuration file needed by FOP*/
    private static final String fopPropertiesPath = "/fopConfiguration.properties";

    /*Contains all couplets found in fopPropertiesPath*/
    static private final Properties fopConfiguration = new Properties();
    static {
        // Set up the XSLT Transformator that should be used.
        // We need this because it is used in the first part of the transformation (XML -> XSL-FO)
        // Invoked here because it is not sure that it has been already set in XMLArtifactTransformer.
        XMLTransformer.initializeDriver();
        // Load properties
        InputStream propertiesStream = FOPTransformer.class.getResourceAsStream(fopPropertiesPath);
        try {
            fopConfiguration.load(propertiesStream);
//            initializeDriver();
        } catch (IOException ioe) {
            final String message;
            message = new StringBuffer().append("It was not possible to load properties from [")
                    .append(fopPropertiesPath).append("] in ")
                    .append(FOPTransformer.class.getName()).toString();
            log.error(message, ioe);
        }
    }

    public FOPTransformer() {
        initializeDriver();
    }

    public void initializeDriver() { //static
        // Construct driver
        fopDriver = new Driver();
        // Set up Logger
//        log.setLevel(Level.ERROR);
        final Log4JLogger log4JLogger = new Log4JLogger(log);
        fopDriver.setLogger(log4JLogger);
        MessageHandler.setScreenLogger(log4JLogger);
        // Set up renderer
        fopDriver.setRenderer(Driver.RENDER_PDF);
        configDirver();
    }

    /**
     * Sets all needed by driver additional configurations.
     * In most cases this method is invoked when font and language
     * of generated PDFs should be cyrillic (e.g. userconfig.xml file needed by FOP to register
     * a needed bulgarian font). Configures and registers all additional service files as
     * userconfig.xml, .ttf(TrueType font) etc. files needed by FOP.
     */
    private void configDirver() { //static
        final String keyConfigPath = "fopConfiguration.configuration.path";
        final String keyUserConfig = "fopConfiguration.userconfig";
        String fopConfigPath = fopConfiguration.getProperty(keyConfigPath);
        String userConfigFileName = fopConfiguration.getProperty(keyUserConfig);
        log.debug("fopConfigPath = " + fopConfigPath);
        log.debug("userConfigFileName = " + userConfigFileName);
        if (!StringUtils.isEmpty(userConfigFileName) && !StringUtils.isEmpty(fopConfigPath)) {
            File baseDir = new File(System.getProperty("MMCAPP_REAL_PATH") + "/WEB-INF/classes");
//            URL fopConfigURL = this.getClass().getResource(fopConfigRelativePath);
//            final String fopConfigAbsolutePath = fopConfigURL.getPath();
//            File userConfigFile = new File(fopConfigAbsolutePath + userConfigFileName);
            File userConfigFile = new File(baseDir, fopConfigPath + userConfigFileName);
            // Registers the fop configuration file to FOP.
            try {
                Options options = new Options(userConfigFile);
                File fontBase = new File(baseDir, fopConfigPath);
                String fontBaseDir = null;
                try {
                    fontBaseDir = fontBase.toURL().toString();
                } catch (MalformedURLException murle) {
                    murle.printStackTrace();
                }
                Configuration.put("fontBaseDir", fontBaseDir);
                log.debug("fontBaseDir = " + fontBaseDir);
            } catch (FOPException fope) {
                fope.printStackTrace();
            }
        } else {
            final StringBuffer message = new StringBuffer("There is no valid value in ");
            message.append("[").append(fopPropertiesPath).append("]").append(" after the specified keyUserConfig ");
            message.append("[").append(keyUserConfig).append("]");
            throw new IllegalArgumentException(message.toString());
        }
    }

    public ByteArrayOutputStream applyTransformation(Source input, Source xslt) {
        // Set output
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        fopDriver.setOutputStream(out);

        try {
            // Set XSLT in the transformer
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer(xslt);

            //Resulting SAX events (the generated FO) must be piped through to FOP
            Result result = new SAXResult(fopDriver.getContentHandler());

            // Start XSL transformation and FOP processing
            transformer.transform(input, result);
        } catch (TransformerConfigurationException tce) {
            StringBuffer message = new StringBuffer("There is a problem when trying to initialize transformer.");
            log.error(message.toString(), tce);
        } catch (TransformerException te) {
            StringBuffer message = new StringBuffer("There is a problem when trying to process the transform.");
            log.error(message.toString(), te);
        } finally {
            try {
                out.close();
            } catch (IOException ioe) {
                StringBuffer message = new StringBuffer("There was problem when invoking OutputStream.close()!");
                log.error(message.toString(), ioe);
            }
        }

        return out;
    }
}
