package org.mmc.model.transformers;

import org.apache.log4j.Logger;
import org.mmc.utils.StringUtils;

import javax.xml.transform.*;
import javax.xml.transform.stream.StreamResult;
import java.util.Properties;
import java.io.IOException;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;

/**
 * Created on: 2007-10-18
 *
 * @author Stoyo Venkov
 */
public class XMLTransformer implements BaseTransformer {

    private static final Logger log = Logger.getLogger(XMLTransformer.class.getName());
    private static final String TRANSFORMER_KEY = "transformer";
    private static final Properties systemTransformProprties = new Properties();
    private static final String systemTransformPropertiesPath = "/systemTransform.properties";
    private static TransformerFactory transformerFactory;

    static {
        InputStream propertiesStream = XMLTransformer.class.getResourceAsStream(systemTransformPropertiesPath);
        try {
            systemTransformProprties.load(propertiesStream);
            initializeDriver();
        } catch (IOException ioe) {
            final String message = new StringBuffer()
                    .append("It was not possible to load properties from [")
                    .append(systemTransformPropertiesPath).append("] in ")
                    .append(XMLTransformer.class.getName()).toString();
            log.error(message, ioe);
        }
    }

    public static void initializeDriver() {
        log.info(new StringBuffer()
                .append("Set TransformerFactory used as system property javax.xml.transform.TransformerFactory.\n")
                .append("This message should be displayed once during first subsystem invocation!").toString());
        String transformerImpl = systemTransformProprties.getProperty(TRANSFORMER_KEY);
        String transformerSys = System.getProperty("javax.xml.transform.TransformerFactory");
        // Checks if proper transformer has already been set as System property.
        if (!transformerImpl.equals(transformerSys)) {
            log.debug("TransformerFactory used as system property javax.xml.transform.TransformerFactory ["+
                    transformerImpl+"]");
            if (StringUtils.isEmpty(transformerImpl)) {
                log.error("There is no such TRANSFORMER under key ["+TRANSFORMER_KEY+"] in "
                        +systemTransformPropertiesPath, new IllegalArgumentException());
            }
            System.setProperty("javax.xml.transform.TransformerFactory", transformerImpl);
        }
    }

    public ByteArrayOutputStream applyTransformation(Source input, Source xslt) {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        StreamResult result = new StreamResult(out);
        try {
            final TransformerFactory transformerFactory = getTransformerFactory();
            Transformer transformer = transformerFactory.newTransformer(xslt);
            transformer.transform(input, result);
            return (ByteArrayOutputStream) result.getOutputStream();
        } catch (TransformerException e) {
            final String message = "There is problem when transformation is invoked!";
            log.error(message, e);
            return null;
        }
    }

    /**
     * Instantiates <code>TransformerFactory</code> new instance.
     * @return transformerFactory
     */
    private static TransformerFactory getTransformerFactory() {
        if (transformerFactory == null) {
            try {
                transformerFactory = TransformerFactory.newInstance();
            } catch (TransformerFactoryConfigurationError e) {
                log.error("Could not instantiate TransformerFactory.newInstance()!", e);
            }
            log.debug("New transformerFactory created ["+transformerFactory+"]");
        }
        return transformerFactory;
    }
}
