package org.mmc.web.actions.search;

import org.mmc.web.actions.BaseSecurityAction;
import org.mmc.web.Constants;
import org.mmc.web.forms.search.SearchValidatorForm;
import org.mmc.model.filters.BasePageFilter;
import org.apache.struts.action.*;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import java.util.List;

/**
 * Created on: 2007-9-15
 *
 * @author Stoyo Venkov
 */
public abstract class BaseSearchAction extends BaseSecurityAction {
    protected ActionForward doExecute(ActionMapping actionMapping,
                                      ActionForm actionForm,
                                      HttpServletRequest httpServletRequest,
                                      HttpServletResponse httpServletResponse)
            throws Exception {

        String filterOp = httpServletRequest.getParameter(Constants.OPERATION_ATTR);

        if(!Constants.OPERATION_SHOW.equals(filterOp)) {
            BasePageFilter filter = initializeFilter(actionMapping, httpServletRequest);

            handleFiltering(actionMapping, actionForm, httpServletRequest);

            handlePaging(actionMapping, httpServletRequest);

            populateForm(actionForm, filter);

            processSearchResult(filter, httpServletRequest);
        } else {
            prePopulateForm(actionForm);
            httpServletRequest.getSession().setAttribute(getFilterAttributeName(), null);
        }

        return actionMapping.findForward(Constants.FORWARD_SUCCESS);
    }

    /**
     * Creates a filter if one does not exist in the session.
     * @param  mapping   <code>ActionMapping</code>
     * @param  request   <code>HttpServletRequest</code>
     * @return a <code>BasePageFilter</code> object
     */
    protected BasePageFilter initializeFilter(ActionMapping mapping,
                                              HttpServletRequest request) {
        HttpSession session = request.getSession();
        BasePageFilter filter = (BasePageFilter)session.getAttribute(getFilterAttributeName());

        if(null == filter) {
            filter = constructFilter();
            session.setAttribute(getFilterAttributeName(), filter);
        }
        return filter;
    }

    /**
     * Handles search operations. Populates the filter with values
     * of the search fields in the form.
     * @param mapping  <code>ActionMapping</code>
     * @param form     <code>ActionForm</code>
     * @param request  <code>HttpServletRequest</code>
     */
    protected void handleFiltering(ActionMapping mapping, ActionForm form,
                                   HttpServletRequest request) {

        String filterOp = request.getParameter(Constants.OPERATION_ATTR);
        if (Constants.OPERATION_FILTER.equals(filterOp)) {

            ActionErrors errors = null;
            if(form instanceof SearchValidatorForm) {
                errors = form.validate(mapping, request);
                if(!errors.isEmpty()) {
                    saveErrors(request, (ActionMessages) errors);
                }
            }

            if(errors == null || errors.isEmpty()) {
                // populate filter from Struts ActionForm
                HttpSession session = request.getSession();
                BasePageFilter filter = (BasePageFilter)session.getAttribute(getFilterAttributeName());
                populateFilter(filter, form);

                //reset page number
                filter.setPageNumber(0);

                //save filter
               session.setAttribute(getFilterAttributeName(), filter);
            }
        }
    }

    /**
     * Hadles paging operations. Populates the filter with the new page
     * and saves it in the session.
     * @param mapping  <code>ActionMapping</code>
     * @param  request <code>HttpServletRequest</code>
     */
    protected void handlePaging(ActionMapping mapping, HttpServletRequest request) {

        String filterOp = request.getParameter(Constants.OPERATION_ATTR);
        if (Constants.OPERATION_PAGING.equals(filterOp)) {
            //get page operation - previous or next
            String pageOp = request.getParameter(Constants.PAGE_OPERATION_ATTR);
            int pageOffset = 0;
            if (Constants.PAGE_NEXT.equals(pageOp)) {
                pageOffset = 1;
            } else if (Constants.PAGE_PREVIOUS.equals(pageOp)) {
                pageOffset = -1;
            } else {
                return;
            }

            //get filter
            HttpSession session = request.getSession();
            BasePageFilter filter = (BasePageFilter)session.getAttribute(getFilterAttributeName());
           //set new page
            filter.setPageNumber(filter.getPageNumber() + pageOffset);
            //save filter
             session.setAttribute(getFilterAttributeName(), filter);
        }
     }

    /**
     * Process the result of the search action. Updates the filter and
     * saves the result in the request
     * @param filter   <code>BasePageFilter</code>
     * @param request  <code>HttpServletRequest</code>
     */
    protected void processSearchResult(BasePageFilter filter, HttpServletRequest request) {
        List results = exposeDbData(filter);
        int resultSize = results.size();
//        if(resultSize == 0) {
//            appendMessage(request, SharedConstants.MESSAGE_NO_DATA, new Object[0]);
//        }
        if(resultSize > 0) {
            if(resultSize >= filter.getPageSize()) {
                request.setAttribute(Constants.SEARCH_RESULT_ATTR,
                                     results.subList(0, filter.getPageSize()));
            } else {
                request.setAttribute(Constants.SEARCH_RESULT_ATTR, results);
            }
        }
        HttpSession session = request.getSession();
        filter = (BasePageFilter)session.getAttribute(getFilterAttributeName());
        filter.setResultSize(resultSize);
        session.setAttribute(getFilterAttributeName(), filter);

    }

    /**
     * Constructs a custom filter for specific search action.
     * @return  a <code>BasePageFilter</code> object
     */
    protected abstract BasePageFilter constructFilter();

    /**
     * Populates the specfied filter with the values
     * of the search form.
     * @param filter <code>BasePageFilter</code>
     * @param actionForm   <code>ActionForm</code>
     */
    protected abstract void populateFilter(BasePageFilter filter, ActionForm actionForm);

    /**
     * Populates the specified form with values from
     * the filter's properties.
     * @param actionForm   <code>ActionForm</code>
     * @param filter <code>BasePageFilter</code>
     */
    protected abstract void populateForm(ActionForm actionForm, BasePageFilter filter);

    /**
     * Gets the data from DB, satisfied the specified filter.
     * @param filter <code>BasePageFilter</code>
     * @return  the result as <code>List</code>
     */
    protected abstract List exposeDbData(BasePageFilter filter);

    /**
     * The action can populate the form with collections form the database.
     *
     * @param actionForm <code>ActionForm</code>
     */
    protected void prePopulateForm(ActionForm actionForm) {

    }

    /**
     * Gets the filter's session attribute name.
     *
     * @return The name of the session attribute.
     */
    protected String getFilterAttributeName() {
        return Constants.SEARCH_FILTER_ATTR;
    }
}
